/* Copyright (C) 2013-2019 Vincent Forest (vaplv@free.fr)
 *
 * The RSys library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published
 * by the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * The RSys library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with the RSys library. If not, see <http://www.gnu.org/licenses/>. */

#include "clock_time.h"
#include <stdlib.h>

int
main(int argc, char** argv)
{
  struct time start, end, res;
  char dump[512];
  char buf[32];
  int64_t time = 0;
  int64_t i = 0;
  (void)argc, (void)argv;

  CHK(time_current(&start) == &start);
  FOR_EACH(i, 0, INT32_MAX / 64); /* Active wait */
  CHK(time_current(&end) == &end);

  CHK(time_sub(&res, &end, &start) == &res);
  time = time_val(&res, TIME_NSEC);
  CHK(time >= 0);
  CHK(time_val(&res, TIME_USEC) == time / 1000);
  CHK(time_val(&res, TIME_MSEC) == time / 1000000);
  CHK(time_val(&res, TIME_SEC) == time / 1000000000);

  time_dump
    (&res, TIME_SEC|TIME_MSEC|TIME_USEC, NULL, dump, sizeof(dump));
  printf(">>> %s.\n", dump);
  time_dump(&res, TIME_ALL, NULL, dump, sizeof(dump));
  printf(">>> %s.\n", dump);

  CHK(time_add(&res, &res, &res) == &res);
  CHK(time_val(&res, TIME_NSEC) == 2*time);
  time_dump(&res, TIME_ALL, NULL, dump, sizeof(dump));
  printf(">>> %s.\n", dump);

  time = time_val(&res, TIME_NSEC);
  CHK(time_zero(&start) == &start);
  CHK(time_val(&start, TIME_NSEC) == 0);
  CHK(time_add(&res, &res, &start) == &res);
  CHK(time_val(&res, TIME_NSEC) == time);
  CHK(time_add(&start, &res, &start) == &start);
  CHK(time_val(&start, TIME_NSEC) == time);
  CHK(time_sub(&res, time_zero(&start), &res) == &res);
  CHK(time_val(&res, TIME_NSEC) == -time);

  CHK(time_sub(&res, &end, &end) == &res);
  CHK(time_val(&res, TIME_NSEC) == 0);
  time_dump(&res, TIME_ALL, NULL, dump, sizeof(dump));
  printf(">>> %s.\n", dump);
  time_dump(&res, 0, NULL, dump, sizeof(dump));
  printf(">>> %s.\n", dump);

  res.sec = 5;
  res.nsec = 524198207;
  time_dump(&res, TIME_ALL, NULL, buf, sizeof(buf));
  printf(">>> %s.\n", buf);

  return 0;
}
